(*********************************************************************)
(*                         Diy                                       *)
(*                                                                   *)
(*   Jade Alglave, Luc Maranget INRIA Paris-Rocquencourt, France.    *)
(*                                                                   *)
(*  Copyright 2010 Institut National de Recherche en Informatique et *)
(*  en Automatique. All rights reserved. This file is distributed    *)
(*  under the terms of the Lesser GNU General Public License.        *)
(*********************************************************************)

open Printf
open Code

let rev = "$Rev: 3300 $"

module type S = sig
  type edge

  type event =
      { loc : loc ;
        v   : v ;
        dir : dir ; }

  type node = {
      mutable evt : event ;
      mutable edge : edge ;
      mutable next : node ;
      mutable prev : node ;
    } 

(* Re-extract edges out of cycle *)
  val extract_edges : node -> edge list

(* Resolve edge direction and build cycle *)
  val resolve_edges : edge list -> edge list * node

(* Finish edge cycle, adding complete events *)
  val finish : node -> unit

(* Composition of the two more basic steps above *)
  val make : edge list -> edge list * node

(* split cycle amoungst processors *)
  val split_procs : node -> node list list


  (* Return coherence orders *)
  val coherence : node -> (loc * v list list) list
end

module Make(E:Edge.S) :
    S with type edge = E.edge
= struct

type edge = E.edge
open E
  type event =
      { loc : loc ;
        v   : v ;
        dir : dir ; }

  type node = {
      mutable evt : event ;
      mutable edge : edge ;
      mutable next : node ;
      mutable prev : node ;
    }

let debug_dir d = match d with W -> "W" | R -> "R"

let debug_evt e =
  sprintf "%s %s %i" (debug_dir e.dir) e.loc e.v

let debug_edge = pp_edge

let debug_node chan n =
  fprintf chan "%s -%s->"
    (debug_evt n.evt) (debug_edge n.edge)

let debug_cycle chan n =
  let rec do_rec m =
    debug_node chan m ;
    output_char chan '\n' ;
    if m.next != n then do_rec m.next in
  do_rec n ;
  flush chan


let evt_null = { loc="*" ; v=(-1) ; dir=R; }


let rec nil =
  {
   evt = evt_null ;
   edge = Po (Diff,Irr,Irr);
   next = nil ;
   prev = nil ;
  }

let alloc_node e = 
  {
   evt = evt_null ;
   edge = e ;
   next = nil ; 
   prev = nil ;
  }

(* Add a node to non-empty cycle *)

let cons_cycle n c =
  n.next <- c ;
  n.prev <- c.prev ;
  c.prev.next <- n ;
  c.prev <- n ;
  n

let build_cycle =

  let rec do_rec es = match es with
  | [] -> assert false (* Empty cycle is absurd *)
  | [e] ->
      let n = alloc_node e in
      n.next <- n ; n.prev <- n ;
      n
  | e::es ->
      cons_cycle (alloc_node e) (do_rec es) in

  fun es ->
    let c = do_rec es in
    c



let find_node p n =
  let rec do_rec m =
    if p m then m
    else
      let m = m.next in
      if m == n then raise Exit
      else do_rec m in
  do_rec n


let find_edge p = find_node (fun n -> p n.edge)


let find_transition p n =
  let n = find_edge (fun n -> not (p n)) n in
  find_edge p n

let find_edge2 p n =
  let rec do_rec m =
    if p m then m
    else
      let m = m.next in
      if m == n then raise Exit
      else do_rec m in
  do_rec n

let find_transition2 p n =
  let n = find_edge2 (fun n -> not (p n)) n in
  find_edge2 p n

(* Add events in nodes *)

module Env = Map.Make(String)

let locs =
  let t = Array.create 26 "" in
  t.(0) <- "x" ;
  t.(1) <- "y" ;
  t.(2) <- "z" ;
  for k=0 to (26-3)-1 do
    t.(k+3) <- String.make 1 (Char.chr (Char.code 'a' + k))
  done ;
  t

let locs_len = Array.length locs
    
let make_loc n =
  if n < locs_len then locs.(n)
  else Printf.sprintf "x%02i" (n-locs_len)
  
let next_loc (loc0,vs) = make_loc loc0,(loc0+1,vs)

let next_v loc (next_loc,vs) =
  let v0 =
    try Env.find loc vs
    with Not_found -> 1 in
  v0,(next_loc,Env.add loc (v0+1) vs)

and cur_v loc (_,vs as st)=
  let v0 =
    try Env.find loc vs
    with Not_found -> 0 in
  v0,st

let same_loc e = match loc_sd e with
    | Same -> true
    | Diff -> false


let find_prec n = n.prev


        
        
(****************************)
(* Add events in edge cycle *)
(****************************)

(* Put directions into edge component of nodes, for
   easier access *)
let patch_edges n =
  let rec do_rec m = 
    let e = set_src m.evt.dir (set_tgt m.next.evt.dir m.edge) in
    m.edge <- e ;
    if m.next != n then do_rec m.next in
  do_rec n

(* Set directions of events *)

let set_dir n0 =
  let rec do_rec p m =
    let d = match dir_tgt p.edge,dir_src m.edge with
    | Irr,Irr ->
        Warn.fatal "Ambiguous direction %s %s"
          (pp_edge p.edge) (pp_edge m.edge)
    | (Dir d,Irr)|(Irr,Dir d) -> d
    | Dir d1,Dir d2 ->
        if d1=d2 then d1
        else
          Warn.fatal "Impossible direction %s %s"
            (pp_edge p.edge) (pp_edge m.edge) in
    m.evt <- { m.evt with dir=d; } ;
    if m.next != n0 then do_rec p.next m.next in
  do_rec (find_prec n0) n0 ;
  patch_edges n0 ;
  if !Misc.verbose > 1 then begin
    eprintf "DIRECTIONS\n" ;
    debug_cycle stderr n0
  end


(***************************)
(* Set locations of events *)
(***************************)

(* Loc is changing *)
let set_diff_loc st n0 =
  let rec do_rec st p m =
    let loc,st =
      if same_loc p.edge then p.evt.loc,st
      else next_loc st in
    m.evt <- { m.evt with loc=loc; } ;
    if m.next != n0 then do_rec st p.next m.next
    else begin
      if m.evt.loc = n0.evt.loc then
        Warn.fatal "Cannot get changing loc accros %s\n"
          (pp_edge m.edge) ;
      st
    end in
  let p = find_prec n0 in
  assert (not (same_loc p.edge)) ;
  do_rec st p n0

(* Loc is not changing *)
let set_same_loc st n0 =
  let loc,st = next_loc st in
  let rec do_rec m =
    m.evt <- { m.evt with loc=loc; } ;
    if m.next != n0 then do_rec m.next in
  do_rec n0 ;
  st



(* Next write in coherence order, or None *)
let find_next_w n0 =
  let rec find_rec n = match n.edge with
  | Ws _|Fr _
  | Po (Same,_,Dir W) | Fenced (_,Same,_,Dir W) | Dp (Same,Dir W)
    -> Some n.next
  | Rf _|Hat
  | Po (Same,_,Dir R) | Fenced (_,Same,_,Dir R)
    -> if n.next == n0 then None else find_rec n.next
  | Dp _|Ctrl _|Po _|Fenced _-> None in
  find_rec n0

(* Idem, backwards *)
let find_prev_w n0 =
  let rec find_rec n = match n.prev.edge with
  | Ws _|Rf _
  | Po (Same,Dir W,_) | Fenced (_,Same,Dir W,_)
    -> Some n.prev
  | Fr _|Hat
  | Po (Same,Dir R,_) | Fenced (_,Same,Dir R,_)
  | Dp (Same,_)|Ctrl (Same,_)
    -> if n.prev == n0 then None else find_rec n.prev
  | Dp _|Po _|Fenced _|Ctrl _ -> None in
  find_rec n0




(* Set the values of write events *)

  let split_by_loc n =
    let rec do_rec m =
      let r =
        if m.next == n then begin
          assert (m.evt.loc <> m.next.evt.loc) ;
          [[]]
        end else do_rec m.next in
      if m.evt.loc =  m.next.evt.loc then match r with
      | ms::rem -> (m::ms)::rem
      | [] -> assert false
      else [m]::r in
    do_rec n

  let split_one_loc n =
    let rec do_rec m =
      m::
      if m.next == n then []
      else n::do_rec m.next in
    [do_rec n]

  let rec do_set_write_val next = function
    | [] -> ()
    | n::ns ->
        begin match n.evt.dir with
        | W ->
            n.evt <- { n.evt with v = next; } ;
            do_set_write_val (next+1) ns
        | R ->  do_set_write_val next ns
        end

  let set_all_write_val nss =
    List.iter
      (do_set_write_val 1)
      nss
  
  let set_write_v n =
    let nss =
      try 
        let m =
          find_node
            (fun m ->
              m.prev.evt.loc <> m.evt.loc &&
              m.next.evt.loc = m.evt.loc) n in
        split_by_loc m
      with Exit -> try
        let m = 
          find_node
            (fun m -> match m.prev.edge with
            | Fr _|Rf _|Ws _|Hat -> false
            | Po _|Dp _|Ctrl _|Fenced _ -> true) n in
        split_one_loc m
      with Exit -> Warn.fatal "Cannot set write values" in
    set_all_write_val nss
  
  

(* Set the value of read_events, from the ones of write events *)
let set_read_v n0 =
  let rec do_rec n =
    begin match n.evt.dir with
    | W -> ()
    | R ->
        let v = match find_prev_w n with
        | Some w -> w.evt.v
        | None -> match find_next_w n with
          | None -> 0 (* Init value *)
          | Some w -> w.evt.v-1 in
        n.evt <- { n.evt with v=v; }
    end ;
    if n.next != n0 then do_rec n.next in
  do_rec n0

(* zyva... *)      

let finish n =
  let st = 0,Env.empty in
(* Set locations *)
  let sd,n =
    let no =
      try Some (find_edge (fun e -> not (same_loc e)) n)
      with Exit -> None in
    match no with
    | Some n ->
        Diff,
        begin try find_edge same_loc n
        with Exit -> Warn.fatal "This cycle changes location at every step" end
    | None -> Same,n in
  let _st = 
    match sd with
    | Diff -> set_diff_loc st n
    | Same -> set_same_loc st n in
  if !Misc.verbose > 1 then begin
    eprintf "LOCATIONS\n" ;
    debug_cycle stderr n
  end ;    
(* Set write values *)
  set_write_v n ;
  if !Misc.verbose > 1 then begin
    eprintf "WRITE VALUES\n" ;
    debug_cycle stderr n
  end ;    
(* Set load values *)
  set_read_v n ;
  if !Misc.verbose > 1 then begin
    eprintf "READ VALUES\n" ;
    debug_cycle stderr n
  end ;
  ()


(* Re-extract edges, with irelevant directions solved *)
let extract_edges n =
  let rec do_rec m = 
    m.edge::
    if m.next == n then []
    else do_rec m.next in
  do_rec n

let resolve_edges = function
  | [] -> Warn.fatal "No edges at all!"
  | es ->
      let c = build_cycle es in
      set_dir c ;
      extract_edges c,c

let make es =
  let es,c = resolve_edges es in
  finish c ;
  es,c

(*************************)
(* Gather events by proc *)
(*************************)

let same_proc e = get_ie e = Int

let find_start_proc n =
  let n = find_edge (fun n -> not (same_proc n)) n in
  try find_edge same_proc n
  with Exit -> n


let cons_not_nil k1 k2 = match k1 with
| [] -> k2
| _::_ -> k1::k2

let split_procs n =
  let n =
    try find_start_proc n
    with Exit -> Warn.fatal "Cannot split in procs" in
  let rec do_rec m =
    let k1,k2 =
      if m.next == n then begin
        if same_proc m.edge then
          Warn.fatal "%s at proc end" (debug_edge m.edge)
        else
          [],[]
      end else do_rec m.next in
    if same_proc m.edge then
      m::k1,k2
    else
      [m],cons_not_nil k1 k2 in
  let k1,k2 = do_rec n in
  cons_not_nil k1 k2

(****************************)
(* Compute coherence orders *)
(****************************)
let same_proc n m =
  let rec do_rec o =
    if o == m then true
    else match E.get_ie o.edge with
    | Int ->
        assert (o.next != n) ;
        do_rec o.next
    | Ext -> false in
  do_rec n


let rec group_rec x ns = function
  | [] -> [x,List.rev ns]
  | (y,n)::rem ->
      if String.compare x y = 0 then group_rec x (n::ns) rem
      else (x,List.rev ns)::group_rec  y [n] rem

  let group = function
    | [] -> []
    | (x,n)::rem -> group_rec x [n] rem

  let by_loc xvs =
    let r = group xvs in
    let r =  List.stable_sort (fun (x,_) (y,_) -> String.compare x y) r in
    let r =
      List.map
        (fun (x,ns) -> match ns with
        |  [] -> assert false
        | n::_ -> (x,List.map (fun n -> n.evt.v) ns))
        r in
    group r
        
      
      
(* find changing location *)
let find_change n =
  let rec do_rec m =
    if m.evt.loc <> m.next.evt.loc then m.next
    else if m.next == n then 
      Warn.fatal "Cannot start coherence order"
    else do_rec m.next in
  do_rec n
      

let coherence n =
  let n = find_change n in
  let rec do_rec m =
    let k =
      if m.next == n then []
      else do_rec m.next in
    let e = m.evt in
    match e.dir with
    | W -> (e.loc,m)::k
    | R -> k in
  let r = by_loc (do_rec n) in
  List.fold_right
    (fun (loc,ws) k -> match ws with
    | [] -> k
    | [_] -> (loc,ws)::k
    | _ -> assert false)
    r []


end
