(*********************************************************************)
(*                         Diy                                       *)
(*                                                                   *)
(*   Jade Alglave, Luc Maranget INRIA Paris-Rocquencourt, France.    *)
(*                                                                   *)
(*  Copyright 2010 Institut National de Recherche en Informatique et *)
(*  en Automatique. All rights reserved. This file is distributed    *)
(*  under the terms of the Lesser GNU General Public License.        *)
(*********************************************************************)

open Code
open Printf

let rev = "$Rev: 3300 $"


module Make(C:XXXCompile.S) : sig

  module A : Arch.S
  type test
  type edge = C.E.edge

(* Returns resolved edges of test *)
  val extract_edges : test -> edge list

(* Build up test, test structure includes
   name & comment given as first two arguments,
   third argument is the last minute check *)
  type info = (string * string) list

  type check = edge list list -> bool
  val make_test :
      string -> ?com:string -> ?info:info -> ?check:check -> edge list -> test
(* Build test from cycle *)
  val test_of_cycle :
      string -> ?com:string -> ?info:info -> ?check:check -> edge list ->
       C.C.node -> test  
(* Dump the given test *)
  val dump_test : test -> unit
  val dump_test_channel : out_channel -> test -> unit


(* Combination of make_test/dump_test *)
(*
  val compile :
      string -> ?com:string -> ?info:MiscParser.info ->
        ?check:check -> edge list -> unit
*)

end
= struct

(* Config *)
  let nprocs = !Config.nprocs
  let do_observers = !Config.do_observers
  let optcond = !Config.optcond
  let overload = !Config.overload

  module A = C.A

  open C.E
  type edge = C.E.edge
  open C.C

  type info = (string * string) list
  type test =
      {
       name : string ;
       com : string ;
       info : info ;
       edges : edge list ;
       init : A.init ;
       prog : A.pseudo list list ;
       final : A.final ;
     }

  let extract_edges {edges=es} = es

(********)
(* Util *)
(********)

module StringSet = MySet.Make(String)

(***************)
(* Compilation *)
(***************)


let as_cons = function
  | x::xs -> x,xs
  | [] -> assert false

and as_some = function
  | Some x -> x
  | None -> assert false

let show_in_cond =
  if optcond then
    let valid_edge = function
      | Rf _ | Fr _ | Ws _ | Hat -> true
      | Po _ | Fenced _ | Dp _|Ctrl _ -> false in
    (fun n -> valid_edge n.prev.edge || valid_edge n.edge)
  else
    (fun _ -> true)
 
  let add_final_v p r v finals = (A.Reg (p,r),v)::finals

  let add_final p o n finals = match o with
  | Some r ->
      if show_in_cond n then add_final_v p r n.evt.v finals
      else finals
  | None -> finals

  let rec emit_overload st p init ov loc =
    if ov <= 0 then init,[],st
    else
      let loc_ov = sprintf "%s%i" loc ov in
      let _,init,i,st = C.emit_load st p init loc_ov in
      let init,is,st = emit_overload st p init (ov-1) loc in
      init,i@is,st

  let insert_overload n = match n.edge with
  | Po (_,Dir R,Dir (W|R)) -> true
  | _ -> false

(* The previous load is... *)
  type 'a prev_load =
    | No       (* Non-existent or irrelevant *)
    | D of 'a  (* My address depends on it *)
    | C of 'a  (* My execution depends on it *)

  let emit_access ro_prev st p init n =
    let init,ip,st = match overload with
    | Some ov  when insert_overload n ->
        emit_overload st p init ov n.evt.loc
    | _ -> init,[],st in
    let o,init,i,st = match ro_prev with
    | No -> C.emit_access st p init n.evt
    | D r1 -> C.emit_access_dep st p init n.evt r1
    | C r1 -> C.emit_access_ctrl st p init n.evt r1 in
    o,init,ip@i,st

let edge_to_prev_load o e = match o with
| None -> No
| Some r ->
    begin match e with
    | Dp _ -> D r
    | Ctrl _ -> C r
    | _ -> No
    end

let rec compile_proc loc_writes st p ro_prev init ns = match ns with
| [] -> init,[],[]
| n::ns ->
      let o,init,i,st = emit_access ro_prev st p init n in
      let init,is,finals =
        compile_proc loc_writes
          st p (edge_to_prev_load o n.edge)
          init ns in
      init,
      i@(match n.edge with Fenced (fe,_,_,_) -> C.emit_fence fe::is |_ -> is),
      (if StringSet.mem n.evt.loc loc_writes then
        add_final p o n finals
      else finals)

let cons_not_nil xs ys = match xs with
| [] -> ys
| _ -> xs::ys

let rec build_observer st p i x = function
  | [] -> i,[],[]
  | [v] ->
      let r,i,c,st = C.emit_load st p i x in
      i,c,add_final_v p r v []
  | v::vs ->
      let r,i,c,st = C.emit_load st p i x in
      let f = C.emit_fence C.A.strong in
      let i,cs,fs = build_observer st p i x vs in
      i,c@[f]@cs,add_final_v p r v fs

let rec build_observers p i x = function
  | [] -> i,[],[]
  | ([]|[_])::vss -> build_observers p i x vss
  | vs::vss ->
      if do_observers then
        let i,c,f = build_observer A.st0 p i x vs in
        let i,cs,fs = build_observers (p+1) i x vss in
        i,c::cs,f@fs
      else Warn.fatal "Observer"

let rec check_rec p i = function
  | [] -> i,[],[]
  | (x,vs)::xvs ->
      let i,c,f = match vs with
      | []|[[_]] -> i,[],[]
      | [[_;v]] -> i,[],[A.Loc x,v]
      | _ -> build_observers p i x vs in
      let i,cs,fs =
        check_rec (p+List.length c) i xvs in
      i,c@cs,f@fs

let check_writes p i n =
  let cos = C.C.coherence n in
  if !Misc.verbose > 1 then begin
    eprintf "COHERENCE: " ;
    Misc.pp_list stderr ""
      (fun chan (x,vs) ->
        fprintf chan "<%s:%a>" x
          (fun chan ->
            Misc.pp_list chan "|"
              (fun chan ->
                Misc.pp_list chan ","
                  (fun chan -> fprintf chan "%i")))
          vs)
      cos ;
    eprintf "\n%!"
  end ;
  check_rec p i cos

let comp_loc_writes n0 =
  let rec do_rec n =
    let k =
      if n.next == n0 then StringSet.empty
      else do_rec n.next in
    match n.evt.dir with
    | W -> StringSet.add n.evt.loc k
    | R -> k in
  do_rec n0

let compile_cycle ok n =
  Label.reset () ;
  let loc_writes = comp_loc_writes n in
  let rec do_rec p i = function
    | [] -> List.rev i,[],[]
    | n::ns ->
        let i,c,f = compile_proc loc_writes A.st0 p No i n in
        let i,cs,fs = do_rec (p+1) i ns in
        i,c::cs,f@fs in

  let i,c,f = check_writes 0 [] n in
  match C.C.split_procs n with
  | [] -> Warn.fatal "No proc"
  | [_] -> Warn.fatal "One proc"
  | splitted ->
      if nprocs <= 0 || List.length splitted <= nprocs then
        let ess = List.map (List.map (fun n -> n.C.C.edge)) splitted in
        if ok ess then
          let i,cs,fs = do_rec (List.length c) i splitted in
          i,c@cs,f@fs
        else Warn.fatal "Last minute check"
      else  Warn.fatal "Too many procs"


(********)
(* Dump *)
(********)

let get_proc = function
  | A.Loc _ -> -1
  | A.Reg (p,_) -> p

let dump_init chan inits =
  fprintf chan "{" ;
  let rec p_rec q = function
    | [] -> fprintf chan "\n}\n"
    | (left,loc)::rem ->
        let p = get_proc left in
        if p <> q then fprintf chan "\n" else fprintf chan " " ;
        fprintf chan "%s=%s;" (A.pp_location left) loc ;
        p_rec p rem in
  p_rec (-1) inits


let rec dump_pseudo = function
  | [] -> []
  | A.Instruction ins::rem -> A.dump_instruction ins::dump_pseudo rem
  | A.Label (lbl,ins)::rem ->
      sprintf "%s:" lbl::dump_pseudo (ins::rem)
  | A.Nop::rem -> dump_pseudo rem
  | A.Macro _::rem -> assert false

let fmt_cols =
  let rec fmt_col p k = function
    | [] -> k
    | cs::prog ->
        (sprintf "P%i" p::dump_pseudo cs)::
        fmt_col (p+1) k prog in
  fmt_col 0 []

let dump_code chan code =
  let pp = fmt_cols code in
  Misc.pp_prog chan pp

let dump_final chan finals =
  fprintf chan "exists\n" ;
  let pp =
    List.map
      (fun (r,v) -> sprintf "%s=%i" (A.pp_location r) v)
      finals in
  fprintf chan "(%s)\n" (String.concat " /\\ " pp)

let dump_test_channel chan t =
  fprintf chan "%s %s\n" (Misc.pp_arch A.arch) t.name ;
  if t.com <>  "" then fprintf chan "\"%s\"\n" t.com ;
  List.iter
    (fun (k,v) -> fprintf chan "%s=%s\n" k v)
    t.info ;
  dump_init chan t.init ;
  dump_code chan t.prog ;
  dump_final chan t.final ;
  ()

let dump_test ({ name = name } as t) =
  let fname = name ^ ".litmus" in
  Misc.output_protect
    (fun chan -> dump_test_channel chan t)
    fname


let test_of_cycle name ?com ?(info=[]) ?(check=(fun _ -> true)) es c =
  let com = match com with None -> pp_edges es | Some com -> com in
  let init,prog,final = compile_cycle check c in
  { name=name ; info=info; com=com ;  edges = es ;
    init=init ; prog=prog ; final=final ; }
    
let make_test name ?com ?info ?check es =
  try
    if !Misc.verbose > 1 then eprintf "**Test %s**\n" name ;
    let es,c = C.C.make es in
    test_of_cycle name ?com ?info ?check es c
  with
  | Misc.Fatal msg ->
      Warn.fatal "Test %s [%s] failed:\n%s" name (pp_edges es) msg
  

 type check = edge list list -> bool

  let compile name ?com ?check es =
    let t = make_test name ?com ?check es in
    dump_test t


end
