(*********************************************************************)
(*                          Litmus                                   *)
(*                                                                   *)
(*        Luc Maranget, INRIA Paris-Rocquencourt, France.            *)
(*        Susmit Sarkar, University of Cambridge, UK.                *)
(*                                                                   *)
(*  Copyright 2010 Institut National de Recherche en Informatique et *)
(*  en Automatique and the authors. All rights reserved.             *)
(*  This file is distributed  under the terms of the Lesser GNU      *)
(*  General Public License.                                          *)
(*********************************************************************)

module type Config = sig
  val word : Globals.word
  val sync_macro : int option
end

module Make(V:Constant.S)(C:Config) =
  struct
    module A = PPCArch.Make(V)
    open A
    open A.Out
    open Printf

(* Ready for template compilation *)
    let op3regs memo set rD rA rB =
      let memo = match set with
      | SetCR0 -> memo ^"."
      | DontSetCR0 -> memo in
      { empty_ins with
        memo=memo^ " ^o0,^i0,^i1";
        inputs=[rA; rB];
        outputs=[rD]; }

    let op2regsI memo rD rA i =
      { empty_ins with
        memo= sprintf "%s ^o0,^i0,%i" memo i;
        inputs=[rA];
        outputs=[rD]; }

    let justOp memo = { empty_ins with memo=memo ; }

    let tr_ins = match C.word with
    | Globals.W64 -> fun i -> i
    | Globals.W32 ->
        fun i -> match i with
        | Pld (a1,a2,a3) -> Plwz (a1,a2,a3)
        | Pstd (a1,a2,a3) -> Pstw (a1,a2,a3)
        | Pldx (a1,a2,a3) -> Plwzx (a1,a2,a3)
        | Pstdx (a1,a2,a3) -> Pstwx (a1,a2,a3) 
        | _ -> i

    let emit_lbl lbl =
      { memo=sprintf "%s:" (A.Out.dump_label lbl) ; inputs=[] ; outputs=[] ;
        label = Some lbl ; branch=[Next] ; }
    
    let next_label =
      let count = ref 0 in
      fun () ->
        let lbl = sprintf "__L%i" !count in
        incr count ;
        lbl

    let li rD v =
      { empty_ins with
        memo=sprintf "li ^o0,%i" v;
        inputs=[];
        outputs=[rD]; }
      
    let mr rD rS =
      { empty_ins with
        memo="mr ^o0,^i0" ;
        inputs=[rS];
        outputs=[rD]; }

    let bcc cond lbl =
      { memo = sprintf "b%s %s" (pp_cond cond) (A.Out.dump_label lbl) ;
        inputs=[] ; outputs=[] ;
        label=None ; branch=[Next; Branch lbl] ; }

    let jump lbl =
      { memo = sprintf "b %s" (A.Out.dump_label lbl) ;
        inputs=[] ; outputs=[] ;
        label=None ; branch=[Branch lbl] ; }

    let incr r i = op2regsI "addi" r r i
    let decr r i =  incr r (-i)

    let cmpwi rS v = 
      { empty_ins with
        memo=sprintf "cmpwi ^i0,%i" v;
        inputs=[rS];
        outputs=[]; } (* no modeling of cc needed here *)

    let cmpw rA rB =
        { empty_ins with
          memo="cmpw ^i0,^i1" ;
          inputs=[rA;rB];
          outputs=[]; }

    let dcbf rA rB =
      { empty_ins with
        memo = "dcbf ^i0,^i1" ;
        inputs=[rA;rB] ; }

    let stwx rS rA rB =
      { empty_ins with
        memo = "stwx ^i0,^i1,^i2";
        inputs=[rS;rA;rB];
        outputs=[]; }
      
    let emit_sync_macro k =
      let delta = Globals.sync_const  in
      let lbl_loop = next_label () in
      let lbl_out = next_label () in
      li idx 0::
      li ephemeral 1::
      cmpwi max_idx 0::
      bcc Gt lbl_loop::
      justOp "sync"::
      jump lbl_out::
      emit_lbl lbl_loop::
      dcbf idx base::
      justOp "sync"::
      stwx ephemeral idx base::
      incr idx delta::
      cmpw idx max_idx::
      bcc Lt lbl_loop::
      emit_lbl lbl_out::k


    let do_compile_ins ins k = match tr_ins ins with
    | Pmr (rD,rS) -> mr rD rS::k

    | Padd(set,rD,rA,rB) -> op3regs  "add" set rD rA rB::k
    | Psub(set,rD,rA,rB) -> op3regs  "subf" set rD rA rB::k
    | Pand(set,rD,rA,rB) -> op3regs  "and" set rD rA rB::k
    | Por(set,rD,rA,rB) ->  op3regs  "or" set rD rA rB::k
    | Pxor(set,rD,rA,rB) -> op3regs  "xor" set rD rA rB::k
    | Pmull(set,rD,rA,rB) -> op3regs  "mullw" set rD rA rB::k
    | Pdiv(set,rD,rA,rB) -> op3regs  "divw" set rD rA rB::k

    | Paddi(rD,rA,simm) -> op2regsI "addi" rD rA simm::k
    | Pandi(rD,rA,simm) -> op2regsI "andi." rD rA simm::k
    | Pori(rD,rA,simm) -> op2regsI "ori" rD rA simm::k
    | Pxori(rD,rA,simm) -> op2regsI "xori" rD rA simm::k
    | Pmulli(rD,rA,simm) -> op2regsI "mulli" rD rA simm::k

    | Pli(rD,v) -> li rD v::k
    | Pcmpwi (0,rS,v) -> cmpwi rS v::k
    | Pcmpwi (crf,rS,v) ->
        { empty_ins with
          memo=sprintf "cmpwi cr%i,^i0,%i" crf v;
          inputs=[rS];
          outputs=[]; }::k
    | Pcmpw(0,rA,rB) -> cmpw rA rB::k
    | Pcmpw(crf,rA,rB) ->
        { empty_ins with
          memo=sprintf "cmpw cr%i,^i0,^i1" crf;
          inputs=[rA;rB];
          outputs=[]; }::k
    | Pb lbl -> jump lbl::k
    | Pbcc(cond, lbl) -> bcc cond lbl::k
    | Plwz (rD,d,rA) ->
        { empty_ins with
          memo = sprintf "lwz ^o0,%i(^i0)"  d;
          inputs = [rA];
          outputs= [rD]; }::k
    | Pld (rD,d,rA) ->
        { empty_ins with
          memo = sprintf "ld ^o0,%i(^i0)"  d;
          inputs = [rA];
          outputs= [rD]; }::k
    | Plwzx (rD,rA,rB) ->
        begin match rA with
        | A.Ireg A.GPR0 -> (* Yes, it's this way cf. ISA p. 48 *)
            { empty_ins with
              memo = "lwzx ^o0,0,^i0";
              inputs = [rB];
              outputs= [rD]; }::k
        | _ -> 
            { empty_ins with
              memo = "lwzx ^o0,^i0,^i1";
              inputs = [rA;rB];
              outputs= [rD]; }::k
        end
    | Pldx (rD,rA,rB) ->
        begin match rA with
        | A.Ireg A.GPR0 -> (* Yes, it's this way cf. ISA p. 48 *)
            { empty_ins with
              memo = "ldx ^o0,0,^i0";
              inputs = [rB];
              outputs= [rD]; }
        | _ -> 
            { empty_ins with
              memo = "ldx ^o0,^i0,^i1";
              inputs = [rA;rB];
              outputs= [rD]; }
        end::k
    | Pstw(rS,d,rA) ->
        { empty_ins with
          memo = sprintf "stw ^i0,%i(^i1)" d;
          inputs=[rS;rA];
          outputs=[]; }::k
   | Pstd(rS,d,rA) ->
        { empty_ins with
          memo = sprintf "std ^i0,%i(^i1)" d;
          inputs=[rS;rA];
          outputs=[]; }::k
    | Pstwx (rS,rA,rB) ->
        begin match rA with
        | A.Ireg A.GPR0 -> (* ISA p. 53 *)
            { empty_ins with
              memo = "stwx ^i0,0,^i1";
              inputs=[rS;rB];
              outputs=[]; }::k
        | _ -> stwx rS rA rB::k
        end
    | Pstdx (rS,rA,rB) ->
        begin match rA with
        | A.Ireg A.GPR0 -> (* ISA p. 53 *)
            { empty_ins with
              memo = "stdx ^i0,0,^i1";
              inputs=[rS;rB];
              outputs=[]; }::k
        | _ ->
            { empty_ins with
              memo = "stdx ^i0,^i1,^i2";
              inputs=[rS;rA;rB];
              outputs=[]; }::k
        end
    | Psync ->
        begin match C.sync_macro with
        | None  ->justOp "sync"::k
        | Some _ -> emit_sync_macro k
        end
    | Plwsync -> justOp "lwsync"::k
    | Pisync -> justOp "isync"::k
    | Peieio -> justOp "eieio"::k
    | Pdcbf (rA,rB) ->
        begin match rA with
        | A.Ireg A.GPR0 -> (* Yes, it's this way cf. ISA p. 415 *)
            { empty_ins with
              memo = "dcbf 0,^i0" ;
              inputs=[rB] ; }
        | _ -> dcbf rA rB
        end::k

    | Plwarx (rT,rA,rB) ->
        begin match rA with
        | A.Ireg A.GPR0 ->
            { empty_ins with
              memo = "lwarx ^o0,0,^i0" ;
              inputs=[rB] ;
              outputs=[rT] ; }
        | _ ->
            { empty_ins with
              memo = "lwarx ^o0,^i0,^i1" ;
              inputs=[rA; rB] ;
              outputs=[rT] ; }
        end::k
    | Pstwcx (rS,rA,rB) ->
        begin match rA with
        | A.Ireg A.GPR0 ->
            { empty_ins with
              memo = "stwcx. ^i0,0,^i1" ;
              inputs=[rS;rB] ;
              outputs=[] ; }
        | _ ->
            { empty_ins with
              memo = "stwcx. ^i0,^i1,^i2" ;
              inputs=[rS ;rA; rB] ;
              outputs=[] ; }
        end::k

    let extract_addrs _ins = StringSet.empty

    let compile_ins ins = do_compile_ins ins
  end
