(*********************************************************************)
(*                          Litmus                                   *)
(*                                                                   *)
(*        Luc Maranget, INRIA Paris-Rocquencourt, France.            *)
(*        Susmit Sarkar, University of Cambridge, UK.                *)
(*                                                                   *)
(*  Copyright 2010 Institut National de Recherche en Informatique et *)
(*  en Automatique and the authors. All rights reserved.             *)
(*  This file is distributed  under the terms of the Lesser GNU      *)
(*  General Public License.                                          *)
(*********************************************************************)

(******************************)
(* Test configuration control *)
(******************************)

(* OS *)
type target_os = Linux | Mac |AIX | AIX5
(* type of barrier *)
type barrier = User | Pthread | NoBarrier
(* how threads are started/joined *)
type launch = Fixed | Changing
type memory = Direct | Indirect
type word = W32 | W64

type config =
    { gcc_opts : string ;
      barrier : barrier ;
      launch : launch ;
      size_of_test : int ;
      number_of_run : int ;
      memory : memory ;
      safer : bool ;
      preload : bool ;
      os : target_os ;
      word : word ;
      avail : int option ; }

let default =
   { gcc_opts = "" ;
    barrier = User ;
    launch = Changing ;
    size_of_test = 100000 ;
    number_of_run = 10 ;
    memory = Indirect ;
    safer = true ;
    preload = true ;
    os = Linux ;
    word = W32 ;
    avail = None ; }


let default_x86 = { default with gcc_opts = "-fomit-frame-pointer -O2" }
let default_ppc = { default with gcc_opts = "-O" }

let config = ref None


let mod_config = ref (fun cfg -> cfg)

let get_config arch =
  let cfg = match !config with
  | None ->
      begin match arch with
      | Misc.X86 -> default_x86
      | Misc.PPC|Misc.ARM -> default_ppc
      end
  | Some cfg -> cfg in
  !mod_config cfg


let replace_config f =
  let g = !mod_config in
  mod_config := (fun cfg -> f (g cfg))

open Printf

let dump_def chan var x =
  fprintf chan "#ifndef %s\n" var ;
  fprintf chan "#define %s %s\n" var x ;
  fprintf chan "#endif\n" ;
  ()

let default_size () = default.size_of_test

let set_size  n = replace_config (fun cfg -> { cfg with size_of_test=n ; })

let default_run () = default.number_of_run

let set_run n = replace_config (fun cfg -> { cfg with number_of_run=n; })

let dump_bench arch chan =
  let config = get_config arch in
  dump_def chan "SIZE_OF_TEST" (string_of_int config.size_of_test) ;
  dump_def chan "NUMBER_OF_RUN" (string_of_int config.number_of_run) ;
  dump_def chan "N_EXE"
    (match config.avail with
    | None -> "1"
    | Some n -> sprintf "(%i < N ? 1 : %i / N)" n n) ;
  ()
  

let set_gcc_opts opts =
  replace_config
    (fun cfg -> { cfg with gcc_opts = opts ; })

and get_gcc_opts arch = (get_config arch).gcc_opts

let default_gcc_opts arch = match arch with
| Misc.X86 -> default_x86.gcc_opts
| Misc.PPC -> default_ppc.gcc_opts
| Misc.ARM -> assert false

let set_barrier tag =
  let b = match tag with
  | "user" -> User
  | "pthread" -> Pthread
  | "none" -> NoBarrier
  | s -> raise (Arg.Bad ("incorrect barrier tag: "^s)) in
  replace_config (fun cfg -> { cfg with barrier = b })

and get_barrier arch = (get_config arch).barrier

let do_pp_barrier = function
| User -> "user"
| Pthread -> "pthread"
| NoBarrier -> "none"

let pp_barrier arch = do_pp_barrier (get_barrier arch)
let default_barrier () = do_pp_barrier default.barrier

(* OS *)
let get_target_os arch = (get_config arch).os

let set_os tag =
  let os = match String.lowercase tag with
  | "linux" -> Linux
  | "mac" -> Mac
  | "aix" -> AIX
  | "aix5" -> AIX5
  | _ -> raise (Arg.Bad ("incorrect os tag: "^tag)) in
  replace_config (fun cfg -> { cfg with os = os; })

let get_word arch = (get_config arch).word

let do_pp_os = function
  | Linux -> "linux"
  | Mac -> "mac"
  | AIX -> "aix"
  | AIX5 -> "aix5"

let default_os () = do_pp_os default.os

let set_word tag =
  let w = match String.lowercase tag with
  | "w32" -> W32
  | "w64" -> W64
  | _ -> raise (Arg.Bad ("incorrect word tag: "^tag)) in
  replace_config (fun cfg -> { cfg with word = w; })

let get_word arch = (get_config arch).word

let do_pp_ws = function
  | W32 -> "w32"
  | W64 -> "w64"

let default_ws () = do_pp_ws default.word

(* Launch technique *)
let set_launch tag =
  let l = match tag with
  | "fixed" -> Fixed
  | "changing" -> Changing
  | s -> raise (Arg.Bad ("incorrect launch tag: "^s)) in
  replace_config (fun cfg -> { cfg with launch = l })

and get_launch arch = (get_config arch).launch

let do_pp_launch = function
| Fixed -> "fixed"
| Changing -> "changing"

let pp_launch arch = do_pp_launch (get_launch arch)

let default_launch () = do_pp_launch default.launch

(* Memory *)
let set_memory tag = 
 let l = match tag with
  | "direct" -> Direct
  | "indirect" -> Indirect
  | s -> raise (Arg.Bad ("incorrect mem tag: "^s)) in
  replace_config (fun cfg -> { cfg with memory = l })

and get_memory arch = (get_config arch).memory

let do_pp_memory = function
| Direct -> "direct"
| Indirect -> "indirect"

let pp_memory arch = do_pp_memory (get_memory arch)
let default_memory arch = do_pp_memory default.memory


(* Safer mode *)
let set_safer v = replace_config (fun cfg -> { cfg with safer=v })
let get_safer arch = (get_config arch).safer
let default_safer () = default.safer

(* Preload of shared globals by each proc *)
let set_preload v = replace_config (fun cfg -> { cfg with preload=v })
let get_preload arch = (get_config arch).preload
let default_preload () = default.preload

(* Flood avail procs *)
let is_flood arch = match (get_config arch).avail with
| None -> false
| Some _ -> true

let get_avail arch = (get_config arch).avail
and set_avail n = replace_config
    (fun cfg ->
      if n > 0 then { cfg with avail=Some n }
      else { cfg with avail=None })

(* Technique for parallel testing on big machines *)
type para = Shell | Self

let dp = Self
let para = ref dp

let set_para tag =
  let p = match tag with
  | "self" -> Self
  | "shell" -> Shell
  | _ ->  raise (Arg.Bad ("incorrect para tag: "^tag)) in
  para := p

let get_para () = !para

let do_pp_para = function
  | Self -> "self"
  | Shell -> "shell"

let pp_para () = do_pp_para !para
let default_para () = do_pp_para dp

(* Do not produce tests with more than <limit> threads *)

let limit = ref false
let set_limit b = limit := b
and get_limit () = !limit

(* bench file (defines number of iterations) *)
let bench = ref None
let set_bench s = bench := Some s
let get_bench () = !bench


(***********************)
(* PPC Sync Macro      *)
(***********************)

let sync_const = 128

let sync_macro_number_times = ref None

let set_sync_macro n =  sync_macro_number_times:= Some n
let get_sync_macro () = !sync_macro_number_times


(***********************)
(* Timing data control *)
(***********************)

(**
 * Brief description of issues in timing litmus runs:
 *
 * Litmus tests are done in blocks of (default) 200 runs, times
 * restarting the threads (default) 1000 times. Each block is attempted
 * with as less synchronization as possible, while the restarts take
 * care of (hopefully) resynchronization.
 * 
 * Method 1: Use processor-local Time Stamp Counter (TSC)
 * ------------------------------------------------------
 * This is an extremely high-res counter (measures in CPU cycles) of time.
 * Very processor-dependent, on X86, there is the RDTSC instruction. On
 * PowerPC, there is an untested sequence of instructions with similar 
 * effect.
 * 
 * This does not do any synchronization, and, (at least on X86) is 
 * out-of-order, thus, rdtsc can return even when preceding stores (eg) have
 * not completed. This makes it less useful as a precise timing, but has
 * value in looking at individual tests (even within blocks).
 * 
 * Note that TSC is not guaranteed to be unique across different cores, so
 * values are only meaningful if on same processor. If there is an intervening
 * preemption by scheduler, all bets are off. With that caveat, it should
 * monotonically increase in time (ignoring 64-bit wraparound issues).
 * 
 * Method 2: Use the POSIX function clock_gettime.
 * -----------------------------------------------
 * This is a POSIX function, implemented within Linux [Note1] in
 * all modern systems. The CLOCK_MONOTONIC is guaranteed to return
 * monotonically increasing values, regardless of calling processor.
 * The cost of this is a system call. On tested systems, it has nanosecond 
 * resolution.
 * 
 * Because of the syscall, this has too much synchronization to be 
 * done within blocks. We perform these calls at the beginning and end of
 * each block, per thread. Thus information (at nanosecond resolution) is
 * made available at the block level.
 * 
 * Note that this requires -lrt (real-time library?) to gcc
 * 
 * [Note1]: Googling suggests AIX supports it as well, untested. MacOSX 
 * DOES NOT, apparently. See here for a possible workaround:
 * http://www.wand.net.nz/~smr26/wordpress/2009/01/19/monotonic-time-in-mac-os-x/ 
 *)
let do_tsc = ref false
let get_tsc () = !do_tsc
let set_tsc b = do_tsc := b

let do_clocktime = ref false
let get_clocktime () = !do_clocktime
let set_clocktime b = do_clocktime := b

(* Profiling with machine time stamps *)
let do_profiling = ref false
let get_profiling () = !do_profiling
let set_profiling b = do_profiling := b


(*******************)
(* Speedcheck mode *)
(*******************)

let ds = false
let speedcheck = ref ds
let set_speedcheck b = speedcheck := b
let get_speedcheck () = !speedcheck
let default_speedcheck () = ds

(************)
(* run name *)
(************)

let runname = ref "default"
let set_runname s = runname := s
let get_runname () = !runname

(*********************)
(* Cross compilation *)
(*********************)

let cross = ref None
let set_cross b  = cross := Some b
and is_cross () = match !cross with
| Some _ -> true
| None -> false
and get_cross () = match !cross with
| Some s -> s
| None -> assert false



