(*********************************************************************)
(*                          Litmus                                   *)
(*                                                                   *)
(*        Luc Maranget, INRIA Paris-Rocquencourt, France.            *)
(*        Susmit Sarkar, University of Cambridge, UK.                *)
(*                                                                   *)
(*  Copyright 2010 Institut National de Recherche en Informatique et *)
(*  en Automatique and the authors. All rights reserved.             *)
(*  This file is distributed  under the terms of the Lesser GNU      *)
(*  General Public License.                                          *)
(*********************************************************************)

(* Run a test and collect statistics *)
open Misc
open Globals
open MiscParser
open Printf
open MyName
open MySys

(* Text quotation for cross-compilation *)
let open_quote chan =
  if is_cross() then output_line chan "cat <<'EOF'"
and close_quote chan =
  if is_cross() then output_line chan "EOF"


let target_os arch =
  if is_cross() then
    get_target_os arch
  else begin
    if Sys.file_exists "/mach_kernel"
    then begin 
      Warn.warn "OS: Darwin" ;
      Mac
    end else begin
      Warn.warn "OS: Linux" ;
      Linux
    end
  end

let get_gcc_opts arch =
  let std_opts =
    "-Wall -std=gnu99 " ^ 
    Globals.get_gcc_opts arch in
  match target_os arch with
  | Mac -> std_opts
  | Linux|AIX -> std_opts ^ " -pthread"
  | AIX5 -> std_opts ^ " -pthread -maix64"

let get_link_opts arch =
  match target_os arch with
  | Mac -> ""
  | Linux|AIX|AIX5 ->
      if  get_clocktime () || get_profiling () then " -lrt"
      else ""
  
let report_parameters arch chan =
  open_quote chan ;
  fprintf chan "Revision %s\n" Config.rev ;
  output_line chan "Parameters" ;
  dump_bench arch chan ;
  fprintf chan "/* gcc options: %s */\n" (get_gcc_opts arch) ;
  fprintf chan "/* barrier: %s */\n" (pp_barrier arch) ;
  fprintf chan "/* launch: %s */\n" (pp_launch arch) ;
  fprintf chan "/* memory: %s */\n" (pp_memory arch) ;
  fprintf chan "/* safer: %b */\n" (get_safer arch) ;
  fprintf chan "/* preload: %b */\n" (get_preload arch) ;
  fprintf chan "/* para: %s */\n" (pp_para ()) ;
  fprintf chan "/* speedcheck: %b */\n" (get_speedcheck ()) ;
  begin match get_avail arch with
  | Some n -> fprintf chan "/* proc used: %i */\n" n
  | None -> ()
  end ;
  close_quote chan ;
  ()

let report_machine arch chan =
(* Machine information *)
  if not (is_cross ()) then begin
    output_string chan "Machine:"  ;
    exec "hostname" (output_line chan) ;
    begin match target_os arch with
    | Linux -> cat "/proc/cpuinfo" (output_line chan)
    | Mac -> exec "system_profiler SPHardwareDataType" (output_line chan)
    | AIX|AIX5 -> ()
    end
  end ;
(* Parameters of test *)
  report_parameters arch chan

module Make(T:Test.S) =
  struct
    module A = T.A
    module C = T.C

(* Parallelism by shell script *)
    let do_shell = match get_para () with
    | Shell -> true
    | Self -> false



    let gcc () =
      assert (not (is_cross ())) ;
      let gcc_opts = get_gcc_opts A.arch in
      let link_opts = get_link_opts A.arch in
      sprintf "gcc %s%s" gcc_opts link_opts

    let gcc2as () =
      assert (not (is_cross ())) ;
      let gcc_opts = get_gcc_opts A.arch in
      sprintf "gcc %s" gcc_opts 

    let compile_test name source =
      let sX = outname name ".exe" in
      if not (is_cross ()) then begin
        let com = sprintf "%s -O2 -o %s utils.c outs.c %s" (gcc()) sX source in
        exec_stdout com
      end ;
      sX

    let assemble_test chan name source =
      let sS = outname name ".s" in
      if is_cross() then
        let com = sprintf "grep _litmus_ %s" sS in
        output_line chan com
      else begin
        let com = sprintf "%s -S -o %s %s"
            (gcc2as()) sS source in
        exec_stdout com ;
        grep chan "_litmus_" sS  ;
        output_line chan ""        
      end ;
      sS

    let run_test chan t name sX =
      let com =
        sprintf "%s %s $LITMUSOPTS" sX (if !verbose > 0 then "-v" else "-q") in      
      if is_cross() then
        match get_avail A.arch with
        | None -> output_line chan com
        | Some n ->
            let nprocs = List.length t.MiscParser.prog in
            let kmax = max 1 (n/nprocs) in
            if do_shell then begin
              if kmax > 1 then begin
                let tmp_file = sprintf "$TMPDIR/%i.log" in
                let output_com c =
                  output_string chan c ;
                  for k=1 to kmax do
                    fprintf chan " %s"  (tmp_file k)
                  done ;
                  output_line chan "" in
                for k=1 to kmax do
                  fprintf chan "%s -n 1 > %s &\n" com (tmp_file k)
                done ;
                output_line chan "wait" ;
                output_com "cat" ;
                output_com "rm -f"
              end else
                fprintf chan "%s -n 1\n" com
            end else begin
              fprintf chan "%s\n" com
            end
      else begin
        exec_stdout com ;
        remove sX ;
        ()
      end 

    let dump_atom (loc,v) =
      sprintf "%s=%s" (A.pp_location loc) (A.V.pp_v v)

    let dump_constr chan = ConstrGen.dump_constraints chan dump_atom

    module D = SimpleDumper.Make
        (struct
           module A = A

          type state = A.state
          let dump_state chan st =
            Misc.pp_list chan " "
              (fun chan a ->
                output_string chan (dump_atom a) ;
                output_char chan ';')
              st

          type constr = C.constr
          let dump_constr = dump_constr

          type location = A.location
          let dump_location chan loc =
            output_string chan (A.pp_location loc)
        end)

open ConstrGen 

let report name doc test source chan =
(* Static information *)
  let title = sprintf "%% Results for %s %%" name in
  let nice = String.make (String.length title) '%' in
  open_quote chan ;
  output_line chan nice ;
  output_line chan title ;
  output_line chan nice ;
  D.dump chan doc test ;
(* Actual assembler *)
  output_string chan "Generated assembler\n" ;
  close_quote chan ;
  flush chan ;
(* End of static information *)
  let sS = assemble_test chan name source in
  remove sS ;
(* Compile *)
  let sX = compile_test name source in
(* Run test *)
  flush chan ;
  run_test chan test name sX ;
  output_line chan "" ;
(* Clean, if not in cross compilation mode *)
  if not (is_cross ()) then begin
    remove source
  end ;
  flush chan




let run name chan doc test source =
  report name doc test source chan

end
