(*********************************************************************)
(*                          Litmus                                   *)
(*                                                                   *)
(*        Luc Maranget, INRIA Paris-Rocquencourt, France.            *)
(*        Susmit Sarkar, University of Cambridge, UK.                *)
(*                                                                   *)
(*  Copyright 2010 Institut National de Recherche en Informatique et *)
(*  en Automatique and the authors. All rights reserved.             *)
(*  This file is distributed  under the terms of the Lesser GNU      *)
(*  General Public License.                                          *)
(*********************************************************************)

(***********************************************)
(* Parse a source file for the needs of litmus *)
(* (And then compile test)                     *)
(***********************************************)

open Printf

type answer =
  | Completed of Misc.arch * string * string list
  | Interrupted of Misc.arch * exn
  | Absent of Misc.arch

module Make (A:Arch.S)
    (L:GenParser.LexParse with type instruction = A.pseudo)
    (XXXComp : XXXCompile.S with module A = A) =
  struct
    module P = GenParser.Make(A) (L)
    module T = Test.Make(A)
    module Comp = Compile.Make (T) (XXXComp)
    module S = Skel.Make(T)
    module R = Run.Make(T)
    open GenParser

    let compile utils name in_chan out_chan splitted  =
      try begin
        let parsed = P.parse in_chan splitted
        and doc = splitted.Splitter.name in
        close_in in_chan ;
        let nprocs = List.length parsed.MiscParser.prog
        and avail = match Globals.get_avail A.arch with
        | Some n -> n
        | None -> 1000 in
        if not (Globals.get_limit ()) || nprocs <= avail then begin
          let module Alloc = SymbReg.Make(A) in
          let allocated = Alloc.allocate_regs parsed in
          let compiled = Comp.compile allocated in
          let source =
            let src = MyName.outname name ".c" in
            if Globals.is_cross () then Tar.outname src
            else src in
          let () = Misc.output_protect  (S.dump doc compiled) source in
          let utils =
            match utils with
            | [] ->
                let module O = ObjUtil.Make(A) in
                O.dump ()
            | _ -> utils  in
          R.run name out_chan doc allocated source ;
          Completed (A.arch,source,utils)
        end else begin
          Warn.warn_always
            "Test %s not performed, too many procs" doc.Name.name ;
          Absent A.arch
        end
      end with e -> Interrupted (A.arch,e)

  end


let from_chan fst name in_chan out_chan =  
(* First split the input file in sections *)
  let arch,splitted =  Splitter.split name in_chan in
(* Then call appropriate compiler, depending upon arch *)
  let module V = SymbConstant in
    match arch with
    | Misc.PPC ->
        let module Conf = struct
          let word = Globals.get_word Misc.PPC
          let sync_macro =Globals.get_sync_macro ()
        end in
	let module PPC = PPCArch.Make(V) in
	let module PPCLexParse = struct
	  type instruction = PPC.pseudo
	  type token = PPCParser.token

	  let lexer = PPCLexer.token
	  let parser = PPCParser.main
	end in
        let module X = Make
            (PPC) (PPCLexParse) (PPCCompile.Make(V)(Conf)) in
        X.compile fst name in_chan out_chan splitted
    | Misc.X86 ->
	let module X86 = X86Arch.Make(V) in
	let module X86LexParse = struct
	  type instruction = X86.pseudo
	  type token = X86Parser.token

	  let lexer = X86Lexer.token
	  let parser = X86Parser.main
	end in
        let module X = Make
            (X86) (X86LexParse) (X86Compile.Make(V)) in
        X.compile fst name in_chan out_chan splitted
    | Misc.ARM ->  Warn.fatal "ARM not implemented"


let from_file fst name out_chan =
  Misc.input_protect
    (fun in_chan -> from_chan fst name in_chan out_chan) name

open MySys

let report_failure name msg chan =
  let title = sprintf "!!! Test %s failed !!!" name in
  let warn = String.make (String.length title) '!' in
  output_line chan  warn ;
  output_line chan  title ;
  output_line chan  warn ;
  fprintf chan "Reported: %s\n" msg ;
  output_line chan ""

let from_files names =
  let out_chan =
    if Globals.is_cross() then
      open_out (Tar.outname (MyName.outname "run" ".sh"))
    else stdout in 
  if Globals.is_cross () then begin
    output_line out_chan "date" ;
    output_line out_chan "LITMUSOPTS=\"$@\"" ;
    begin match Globals.get_para () with
    | Globals.Shell ->
        output_line out_chan "TMPDIR=${TMPDIR:-/var/tmp}" ;
        output_line out_chan "TMPDIR=$TMPDIR/tmp.$$" ;
        output_line out_chan "mkdir $TMPDIR"
    | Globals.Self -> ()
    end
  end ;
  let a,sources,utils =
    List.fold_left
      (fun (_,srcs,fst) name ->
        match from_file fst name out_chan with
        | Completed (a,src,fst) -> a,(src::srcs),fst
        | Absent a -> a,srcs,fst
        | Interrupted (a,e) ->
            let msg =  match e with
            | Misc.Exit -> "None"
            | Misc.Fatal msg 
            | Misc.UserError msg ->
                Pos.pp_pos0 stderr name ;
                Printf.eprintf "%s\n%!" msg ;
                msg
            | e -> raise e  in
            report_failure name msg out_chan ;
            a,srcs,fst)
      (Misc.X86,[],[])  names in
  if (Globals.is_cross ()) then begin
    match Globals.get_para () with
    | Globals.Shell -> output_line out_chan "rmdir $TMPDIR"
    | Globals.Self -> ()
  end ;
  Run.report_machine a out_chan ;
  if (Globals.is_cross ()) then begin
    output_line out_chan "head -1 comp.sh" ;
    output_line out_chan "echo \"LITMUSOPTS=$LITMUSOPTS\"" ;
    output_line out_chan "date"
  end ;
  if (not (Globals.is_cross ())) then begin
    List.iter MySys.remove utils        
  end ;
  flush out_chan ;


  if Globals.is_cross () then begin
    close_out out_chan ;
    let sources = List.map Filename.basename  sources in
    Misc.output_protect
      (fun chan ->
        let gcc_opts = Run.get_gcc_opts a in
        fprintf chan "GCCOPTS=\"%s\"\n" gcc_opts ;
        let link_opts = Run.get_link_opts a in
        fprintf chan "LINKOPTS=\"%s\"\n" link_opts ;
        fprintf chan "/bin/rm -f *.exe *.s\n" ;
        List.iter
          (fun s ->
            if Filename.check_suffix s ".c" then
              fprintf chan "gcc $GCCOPTS -O2 -c %s\n"
                (Filename.basename s))
          utils ;
        List.iter
          (fun src ->
            let exe = Filename.chop_extension src ^ ".exe" in
            let outs = "outs.o" in
(* Moderate parallelism *)
            fprintf chan
              "gcc $GCCOPTS $LINKOPTS -o %s utils.o %s %s &\n" exe outs src ;
            let srcS = Filename.chop_extension src ^ ".s" in
            fprintf chan "( gcc $GCCOPTS -S %s && mv %s /var/tmp/%s && grep _litmus_ /var/tmp/%s > %s && /bin/rm /var/tmp/%s )&\n"
              src srcS srcS srcS srcS srcS ;
            fprintf chan "wait\n")
          (List.rev sources))
      (Tar.outname (MyName.outname "comp" ".sh")) ;
    let tar = Globals.get_cross () in
    Tar.tar tar
  end ;
  ()
    

