(*********************************************************************)
(*                          Litmus                                   *)
(*                                                                   *)
(*        Luc Maranget, INRIA Paris-Rocquencourt, France.            *)
(*        Susmit Sarkar, University of Cambridge, UK.                *)
(*                                                                   *)
(*  Copyright 2010 Institut National de Recherche en Informatique et *)
(*  en Automatique and the authors. All rights reserved.             *)
(*  This file is distributed  under the terms of the Lesser GNU      *)
(*  General Public License.                                          *)
(*********************************************************************)

open Globals
open Printf
let pgm = if Array.length Sys.argv > 0 then Sys.argv.(0) else "litmus"


let opts = Arg.align
  [
(* General behavior *)
   "-v", Arg.Unit (fun () -> incr Misc.verbose), " be verbose";
   "-version", Arg.Unit (fun () -> print_endline Config.version; exit 0),
   " show version number and exit";
   "-libdir", Arg.Unit (fun () -> print_endline Config.libdir; exit 0),
   " show installation directory and exit";
   "-o", Arg.String set_cross, "<name> cross compilation to tar file <name>" ;
   "-cross",  Arg.String set_cross, "<name> same as -o";
   "-mach", Arg.String MyName.read_cfg,
   "<name> read configuration file name.cfg";
   "-runname", Arg.String set_runname,
   "<name> set run name (for condition selection)";

(* Test parameters *)
   "-a", Arg.Int set_avail,
     "<n> Run maximal number of tests concurrently for n available cores (default, run one test)";
   "-limit", Arg.Bool set_limit,
   sprintf
     "<bool> only run tests with n threads or less (default %b)"
     (get_limit ());
   "-s", Arg.Int set_size,
   sprintf
     "<n> size of test (default %i)"
     (default_size ()) ;
   "-r", Arg.Int set_run,
   sprintf
     "<n> number of runs (default %i)"
     (default_run ()) ;

(* Modes *)
   "-barrier",
   Arg.String set_barrier,
   sprintf
     "(user|pthread|none) set type of barriers (default %s)"
     (default_barrier ()) ;
     

   "-launch",
   Arg.String set_launch,
   sprintf
     "(changing|fixed) set type of phread lauch (default %s)"
     (default_launch ()) ;
   
   "-mem",
   Arg.String set_memory,
   sprintf
     "(direct|indirect) set type of memory access (default %s)"
     (default_memory ()) ;

   "-preload", Arg.Bool set_preload,
   sprintf "<bool> preload mode (default %b)" (default_preload ()) ;

   "-safer",
   Arg.Bool set_safer,
   sprintf "<bool> safer mode (default %b)" (default_safer ()) ;
     
   "-para",
   Arg.String set_para,
   sprintf
     "(self|shell) in cross compilation scripts, technique to run several instances: internally (as threads) or externally (as processes). Default %s"
     (default_para ()) ;

   "-speedcheck",Arg.Bool set_speedcheck,
   sprintf
     "<bool> stop test as soon as condition is settled (default %b)"
     (default_speedcheck ()) ;

   "-ccopts", Arg.String set_gcc_opts,
   sprintf "<flags> set gcc compilation flags (default X86=\"%s\", PPC=\"%s\")"
     (default_gcc_opts Misc.X86)
     (default_gcc_opts Misc.PPC) ;

(* Specify target machine *)
   "-os",
   Arg.String set_os,
   sprintf
     "(linux|mac|aix|aix5) target operating system (default %s)"
     (default_os ()) ;

   "-ws", Arg.String set_word,
   sprintf
     "(w32|w64) word size, only works for PPC (default %s)"
     (default_ws ()) ;

(* Misc *)

(*******************)
(* Various timings *)
(*******************)

   "-tsc", Arg.Bool set_tsc, "<bool> get timing info from timestamp counter" ;
   "-clocktime", Arg.Bool set_clocktime, "<bool> get timing info from POSIX clock_gettime function" ;
   "-prof", Arg.Bool set_profiling, "<bool> profile with timestamp counter" ;

(* Bizarre *)
   "-syncmacro", Arg.Int set_sync_macro, "<n> undocumented" ;
  ]


let usage = sprintf   "Usage: %s [opts]* filename" pgm

let sources = ref []

let () = Arg.parse opts (fun s -> sources := s :: !sources) usage 

let sources = Misc.expand_argv !sources

let () =
  try
    Top.from_files sources ; exit 0
  with Misc.Fatal msg ->
    eprintf "Fatal error: %s\n%!" msg ;
    exit 2

